/*----------------------------------------------------------------------
  File    : CoCoGUI.java
  Contents: CoCoNAD toolbox (with pattern spectrum filtering
                             and  pattern set reduction)
  Author  : Christian Borgelt
  History : 2013.11.25 file created
            2013.11.27 pattern spectrum viewer added
            2013.11.28 pattern set viewer added
            2013.11.29 train   set viewer added
            2013.12.10 creation of Executor object added
            2014.03.07 inputs for pattern spectrum estimation added
            2014.08.11 item sequence mining added (respect item order)
            2014.10.14 reading/writing of sequence flag to config. file
            2014.10.20 trimming of spaces from number arguments added
            2014.10.23 changed from LGPL license to MIT license
            2015.08.12 adapted to modified pattern set reduction
            2016.10.05 adapted to extended JNICoCo implementation
            2016.11.04 adapted to fim modules (Pattern, PatternSet etc.)
            2018.11.14 replaced some deprecated functions
            2019.10.11 bug in pmod specification fixed (JNICoCo.NONE)
----------------------------------------------------------------------*/
package coconad;

import java.io.IOException;
import java.io.FileNotFoundException;
import java.io.File;
import java.io.FileReader;
import java.io.FileWriter;
import java.awt.Component;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.ItemEvent;
import java.awt.event.ItemListener;
import javax.swing.JOptionPane;
import javax.swing.JFileChooser;
import javax.swing.JTextField;
import javax.swing.JSpinner;
import javax.swing.JCheckBox;
import javax.swing.JComboBox;
import javax.swing.AbstractButton;
import javax.swing.JFileChooser;

import util.TableReader;
import util.Executor;
import util.ObjExecutor;
import table.Table;
import table.Column;
import table.NominalType;
import table.MetricType;
import table.IntegerType;
import table.RealType;
import table.TableView;
import dialog.TabbedGUI;
import dialog.DialogPanel;
import dialog.FormatPanel;
import dialog.AboutPanel;
import fim.PatternSet;
import fim.PatternView;

/*--------------------------------------------------------------------*/
/** Class for a user interface to the coconad library.
 *  @author Christian Borgelt
 *  @since  2013.11.25 */
/*--------------------------------------------------------------------*/
public class CoCoGUI extends TabbedGUI
  implements ActionListener, ItemListener {

  /*------------------------------------------------------------------*/
  /*  constants                                                       */
  /*------------------------------------------------------------------*/
  private static final long   serialVersionUID = 0x00010013L;
  /** the version string */
  public  static final String VERSION          = "1.19 (2019.10.11)";

  /** the names of the target pattern types */
  private static final String[] trgnames = {
    "item sets", "permutations", "sequences" };

  /** the names of the target pattern subtypes */
  private static final String[] subnames = {
    "frequent/all", "closed", "maximal" };

  /** the codes of the target pattern types */
  private static final int[] trgcodes = {
    PSFnPSR.ITEMSET, PSFnPSR.PERMUTATION, PSFnPSR.SEQUENCE };

  /** the codes of the target pattern subtypes */
  private static final int[] subcodes = {
    PSFnPSR.FREQUENT, PSFnPSR.CLOSED, PSFnPSR.MAXIMAL };

  /** the names of the surrogate generation methods */
  private static final String[] surnames = {
    "none (no pattern spectrum filtering)",
    "read pattern spectrum from file",
    "estimate pattern spectrum",
    "identity (keep original data)",
    "event/spike time randomization",
    "event/spike dithering/displacement",
    "(spike) train shifting/dithering",
    "sampling from a kernel estimate",
    "dithered item permutation",
    "dithered blocked item permutation" };

  /** the codes of the surrogate generation methods */
  private static final int[] surcodes = {
    PSFnPSR.NONE,     PSFnPSR.READ,    PSFnPSR.ESTIMATE,
    PSFnPSR.IDENTITY, PSFnPSR.RANDOM,  PSFnPSR.DITHER,   PSFnPSR.SHIFT,
    PSFnPSR.KERNEST,  PSFnPSR.PERMUTE, PSFnPSR.BLOCK };

  /** the names of the probability density functions */
  private static final String[] pdfnames = {
    "rectangular", "triangular", "Gaussian/normal" };

  private static final int[] pdfcodes = {
    PSFnPSR.UNIFORM, PSFnPSR.TRIANG, PSFnPSR.NORMAL };

  /** the names of the pattern set reduction methods */
  private static final String[] psrnames = {
    "none (keep all patterns after filtering)",
    "excess coincidences (zB,cB-cA)",
    "excess coincidences (zB,cB-cA+1)",
    "excess items/neurons (zA-zB+2,cA)",
    "covered events/spikes zA*cA : zB*cB",
    "covered events/spikes (zA-1)*cA : (zB-1)*cB",
    "combined lenient (z, break rejection tie)",
    "combined lenient (z-1, break rejection tie)",
    "combined strict (z, force decision)",
    "combined strict (z-1, force decision)" };

  /** the codes of the pattern set reduction methods */
  private static final char[] psrcodes = {
    PatternSet.KEEP,     PatternSet.COINS0, PatternSet.COINS1,
    PatternSet.ITEMS2,   PatternSet.COVER0, PatternSet.COVER1,
    PatternSet.LENIENT0, PatternSet.LENIENT1,
    PatternSet.STRICT0,  PatternSet.STRICT1 };

  /*------------------------------------------------------------------*/
  /*  instance variables                                              */
  /*------------------------------------------------------------------*/
  /* --- files --- */
  /** the trains file */
  private JTextField        fn_tns;
  /** the beginning of the event/time range */
  private JTextField        beg;
  /** the end       of the event/time range */
  private JTextField        end;
  /** the pattern spectrum file */
  private JTextField        fn_psp;
  /** the pattern/item set output file file */
  private JTextField        fn_out;

  /* --- CoCoNAD parameters --- */
  /** the target pattern type */
  private JComboBox<String> target;
  /** the target pattern subtype for original data */
  private JComboBox<String> suborig;
  /** the target pattern subtype for surrogates */
  private JComboBox<String> subsurr;
  /** the minimum size of an item set */
  private JSpinner          minsize;
  /** the maximum size of an item set */
  private JSpinner          maxsize;
  /** the minimum support of an item set */
  private JSpinner          minsupp;
  /** the width of the time window/maximum distance */
  private JTextField        width;

  /* --- pattern spectrum filtering --- */
  /** the number of surrogate data sets */
  private JSpinner          surcnt;
  /** the surrogate data generation method */
  private JComboBox<String> surr;
  /** the probability density function for the surrogates */
  private JComboBox<String> pdf;
  /** the sigma parameter for the probability density function */
  private JTextField        sigma;
  /** the block size (time interval) for blocked permutations */
  private JTextField        delta;
  /** the item probability dispersion factor */
  private JTextField        alpha;
  /** the number of samples to draw per item set size */
  private JSpinner          smpls;
  /** the seed for the pseudo-random number generator */
  private JTextField        seed;
  /** the components to enable/disable or hide/set visible */
  private Component[]       comps;

  /* --- pattern set reduction --- */
  /** the pattern set reduction method */
  private JComboBox<String> psr;

  /* --- output --- */
  /** the field separator for the pattern spectrum */
  private JTextField        pspsep;
  /** the record header for the item set output */
  private JTextField        header;
  /** the item separator for the item set output */
  private JTextField        itemsep;
  /** the output format for the support information */
  private JTextField        outfmt;

  /*------------------------------------------------------------------*/
  /** Create a CoCoNAD GUI.
   *  <p>The dialog is created as a stand-alone program.
   *  That is, closing or quitting it terminates the program.</p>
   *  @since  2013.11.25 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public CoCoGUI ()
  { this.init(null, true); }

  /*------------------------------------------------------------------*/
  /** Create a CoCoNAD GUI.
   *  @param  owner the component that is to own this dialog
   *  @since  2013.11.25 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public CoCoGUI (Component owner)
  { this.init(owner, false); }

  /*------------------------------------------------------------------*/
  /** Create the user interface.
   *  <p>Following the recommendations in the Java tutorial, the user
   *  interface is created in the "run" method, which is invoked from
   *  the event queue, in order to avoid problems with threads.</p>
   *  @since  2013.11.25 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public void run ()
  {                             /* --- create the user interface */
    DialogPanel tab;            /* current tab of the tabbed pane */

    /* --- create basic GUI --- */
    this.base("CoCoNAD Tools (PSF+PSR)");
    this.addFormatTab(FormatPanel.TRAIN);

    /* --- Files --- */
    tab = this.addTab("Files");
    tab.addLabel("(Spike) trains file:");
    tab.addButton("Select", DialogPanel.MIDDLE).addActionListener(
      new ActionListener () {
      public void actionPerformed (ActionEvent e) {
        CoCoGUI.this.getFileName(CoCoGUI.this.fn_tns); } } );
    tab.addButton("View", DialogPanel.RIGHT).addActionListener(
      new ActionListener () {
      public void actionPerformed (ActionEvent e) {
        CoCoGUI.this.showTrains(CoCoGUI.this.fn_tns); } } );
    this.fn_tns = tab.addFileInput("trains.txt");
    tab.addLabel("(Spike) time range:");
    this.beg    = tab.addNumberInput("", DialogPanel.MIDDLE);
    this.end    = tab.addNumberInput("", DialogPanel.RIGHT);
    tab.addHelp("Only events/spikes in this (time) range are "
               +"considered. If empty,\nthe start time is found "
               +"by rounding the earliest time down to an\ninteger; "
               +"the end time by rounding the latest time up "
               +"to an integer.");
    tab.addFiller(6);

    tab.addLabel("Pattern spectrum file:");
    tab.addButton("Select", DialogPanel.MIDDLE).addActionListener(
      new ActionListener () {
      public void actionPerformed (ActionEvent e) {
        CoCoGUI.this.getFileName(CoCoGUI.this.fn_psp); } } );
    tab.addButton("View", DialogPanel.RIGHT).addActionListener(
      new ActionListener () {
      public void actionPerformed (ActionEvent e) {
        CoCoGUI.this.showSpectrum(CoCoGUI.this.fn_psp); } } );
    this.fn_psp = tab.addFileInput("spectrum.txt");
    tab.addHelp("The pattern spectrum derived from the surrogates "
               +"may be stored\nin a (text) file; or it is read from "
               +"a (text) file instead of generated.");
    tab.addFiller(6);

    tab.addLabel("Pattern output file:");
    tab.addButton("Select", DialogPanel.MIDDLE).addActionListener(
      new ActionListener () {
      public void actionPerformed (ActionEvent e) {
        CoCoGUI.this.getFileName(CoCoGUI.this.fn_out); } } );
    tab.addButton("View", DialogPanel.RIGHT).addActionListener(
      new ActionListener () {
      public void actionPerformed (ActionEvent e) {
        CoCoGUI.this.showPatterns(CoCoGUI.this.fn_out); } } );
    this.fn_out = tab.addFileInput("patterns.txt");
    tab.addHelp("Found patterns (after filtering and "
               +"reduction) are written to this file.");
    tab.addFiller(0);

    /* --- CoCoNAD parameters --- */
    tab = this.addTab("CoCoNAD");
    tab.addLabel("Target pattern type:");
    this.target = tab.addComboBox(CoCoGUI.trgnames);
    this.target.setSelectedIndex(0);
    tab.addLabel("Subtype (original):");
    this.suborig = tab.addComboBox(CoCoGUI.subnames);
    this.suborig.setSelectedIndex(1);
    tab.addLabel("Subtype (surrogates):");
    this.subsurr = tab.addComboBox(CoCoGUI.subnames);
    this.subsurr.setSelectedIndex(0);
    tab.addLabel("Time window width:");
    this.width = tab.addNumberInput("0.003");
    tab.addHelp("Only (groups of) events that can be covered by "
               +"a (sliding) window\nof this width are seen as "
               +"(approximately) synchronous/coincident\nand thus "
               +"may be counted for the item set/pattern support.");
    tab.addLabel("Minimum support:");
    this.minsupp = tab.addSpinner(2, 1, 999999, 1);
    tab.addHelp("Only item sets/patterns with a support of at least "
               +"this number of\n(approximate) "
               +"coincidences/synchronous events are reported.");
    tab.addLabel("Minimum size:");
    this.minsize = tab.addSpinner(2, 1, 999999, 1);
    tab.addLabel("Maximum size:");
    this.maxsize = tab.addSpinner(0, 0, 999999, 1);
    tab.addHelp("Only item sets/patterns with a size in this range "
               +"are reported.\nAn maximum size of 0 means that there "
               +"is no upper limit.");
    tab.addFiller(0);

    /* --- Pattern Spectrum Filtering parameters --- */
    tab = this.addTab("Filtering");
    tab.addLabel("Surrogate data sets:");
    this.surcnt = tab.addSpinner(1000, 1, 999999, 1);
    tab.addHelp("The number of surrogate data sets should be computed "
               +"as\n(an estimate of) the number of signatures "
               +"(i.e. size/support pairs)\nto test divided by the "
               +"desired overall significance level.");
    tab.addLabel("Surrogate data generation method:",DialogPanel.RIGHT);
    this.surr = tab.addComboBox(CoCoGUI.surnames);
    this.surr.setSelectedIndex(2);
    this.surr.addItemListener(this);

    this.comps = new Component[14];
    this.comps[ 0] = tab.addLabel("Density function:");
    this.comps[ 1] = this.pdf = tab.addComboBox(CoCoGUI.pdfnames);
    this.pdf.setSelectedIndex(0);
    this.comps[ 2] = tab.addHelp("For displacements or shifts "
                       +"in the surrogate data generation.");
    this.comps[ 3] = tab.addLabel("Sigma parameter:");
    this.comps[ 4] = this.sigma = tab.addNumberInput("0.005");
    this.comps[ 5] = tab.addHelp("Standard deviation or half the "
                       +"base width of the density function.");
    this.comps[ 6] = tab.addLabel("Block size:");
    this.comps[ 7] = this.delta = tab.addNumberInput("0.03");

    this.comps[ 8] = tab.addLabel("Dispersion factor:");
    this.comps[ 9] = this.alpha = tab.addNumberInput("0.5");
    this.comps[10] = tab.addHelp("Factor with which the item "
                       +"probability dispersion is multiplied.");
    this.comps[11] = tab.addLabel("Samples per size:");
    this.comps[12] = this.smpls = tab.addSpinner(1000, 1, 999999, 1);
    this.comps[13] = tab.addHelp("Number of samples per item "
                       +"set size for probability estimate.");
    tab.addHelp("Time interval for blocked event/spike permutations.");
    for (int k = 0; k < this.comps.length; k++)
      this.comps[k].setVisible(k >= 8);
    this.delta.setEnabled(false);

    tab.addLabel("Random seed:");
    this.seed = tab.addNumberInput("");
    tab.addHelp("If no value is given, "
               +"the current time is used as a seed.");
    tab.addFiller(0);

    /* --- Pattern Set Reduction parameters --- */
    tab = this.addTab("Reduction");
    tab.addLabel("Pattern set reduction method:", DialogPanel.RIGHT);
    this.psr = tab.addComboBox(CoCoGUI.psrnames);
    this.psr.setSelectedIndex(5);
    tab.addHelp("\nA and B refer to two patterns (item sets) "
               +"that are compared,\nwhere B is a subset of A. "
               +"The sizes of A and B (number of items)\nare "
               +"denoted by zA and zB, respectively; the "
               +"support of A and B\n(number of coincidences) "
               +"by cA and cB, respectively.\n\n"
               +"Lenient combination means that both patterns "
               +"may be kept\nif neither criterion prefers one "
               +"pattern to the other, and only\nfor contradicting "
               +"preferences a secondary criterion is invoked.\n"
               +"Strict combination means that a decision "
               +"is always enforced\n(that is, a preference "
               +"is definitely established).\n\n"
               +"Pattern set reduction is not applicable "
               +"for item sequences.");
    tab.addFiller(0);

    /* --- Output --- */
    tab = this.addTab("Output");
    tab.addLabel("Field separator:");
    tab.add(this.pspsep = new JTextField(" "), DialogPanel.RIGHT);
    tab.addHelp("String to separate the fields of a pattern spectrum.");
    tab.addLabel("Item set header:");
    tab.add(this.header = new JTextField(""), DialogPanel.RIGHT);
    tab.addHelp("String that is printed before each item set/pattern.");
    tab.addLabel("Item separator:");
    tab.add(this.itemsep = new JTextField(" "), DialogPanel.RIGHT);
    tab.addHelp("String to separate the items of an item set/pattern.");
    tab.addLabel("Support format:");
    this.outfmt = tab.addTextInput(" (%d)");
    tab.addHelp("The support output format is important "
               +"for the parser of the\nitem set viewer to work "
               +"correctly. Do not change this if you want\nto use "
               +"the viewer. Likewise, do not change the item set "
               +"header\nor the item separator in this case.");
    tab.addFiller(0);

    /* --- About --- */
    this.addTab("About", new AboutPanel(
       "CoCoNAD Tools",
       "Continuous-time Closed Neuron Assembly Detection with\n"
      +"Pattern Spectrum Filtering (PSF) "
      +"& Pattern Set Reduction (PSR)\n\n"
      +"GUI version "     +CoCoGUI.VERSION +", "
      +"JNICoCo version " +JNICoCo.VERSION +"\n"
      +"written by Christian Borgelt\n"
      +"European Centre for Soft Computing\n"
      +"c/ Gonzalo Gutierrez Quiros s/n\n"
      +"33600 Mieres, Asturias, Spain\n"
      +"christian@borgelt.net"));

    /* --- finalize --- */
    this.pack();
    this.selectTab(1);
  }  /* run() */

  /*------------------------------------------------------------------*/
  /** Handle the state change of the surrogate type combo box.
   *  @param  e the change event to react to
   *  @since  2013.12.11 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public void itemStateChanged (ItemEvent e)
  {                             /* --- handle selection changes */
    int i = this.surr.getSelectedIndex();
    if (i == 2) {               /* if to estimate a pattern spectrum */
      for (int k = 0; k < this.comps.length; k++)
        this.comps[k].setVisible(k >= 8); }
    else {                      /* if to generate a pattern spectrum */
      for (int k = 0; k < this.comps.length; k++)
        this.comps[k].setVisible(k <  8); }
    this.surcnt.setEnabled(i > 3);
    this.pdf.setEnabled(i > 3);
    this.sigma.setEnabled(i > 3);
    this.delta.setEnabled(i == 9);
    this.seed.setEnabled(i > 3);
    this.psr.setEnabled(i != 3);
  }  /* itemStateChanged() */

  /*------------------------------------------------------------------*/
  /** Show (spike) trains.
   *  @param  txt the text field containing the file name
   *  @since  2013.11.25 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  protected void showTrains (JTextField txt)
  { this.showTrains(new File(txt.getText())); }

  /*------------------------------------------------------------------*/
  /** Show (spike) trains.
   *  @param  file the file to load the spike trains
   *  @since  2013.11.25 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  protected void showTrains (File file)
  {                             /* --- show a train set */
    double    beg, end;         /* beginning and end of event range */
    TrainView view;             /* created train set viewer */

    view = new TrainView(this, TrainView.LOAD_ITEMS);
    if (!view.loadTrains(file)) return;
    try { beg = Double.parseDouble(this.beg.getText()); }
    catch (NumberFormatException e) { beg = Double.NEGATIVE_INFINITY; }
    try { end = Double.parseDouble(this.end.getText()); }
    catch (NumberFormatException e) { end = Double.POSITIVE_INFINITY; }
    view.setRange(beg, end);    /* set the event range to display */
    view.setVisible(true); view.toFront();
  }  /* showTrains() */

  /*------------------------------------------------------------------*/
  /** Show a pattern spectrum.
   *  @param  txt the text field containing the file name
   *  @since  2013.11.25 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  protected void showSpectrum (JTextField txt)
  { this.showSpectrum(new File(txt.getText())); }

  /*------------------------------------------------------------------*/
  /** Show a pattern spectrum.
   *  @param  file the file to load the pattern spectrum
   *  @return whether loading the spectrum was successful
   *  @since  2013.11.25 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  protected boolean showSpectrum (File file)
  {                             /* --- show a pattern spectrum */
    if (file == null) return false;   /* check for a file name */
    try {                       /* create a reader for the file */
      long t = System.currentTimeMillis();
      System.err.print("reading " +file +" ... ");
      TableReader rdr = this.format.createReader(file);
      Table psp = new Table(file.getPath());
      psp.read(rdr, Table.NOHEADER);
      rdr.close();              /* read the pattern spectrum */
      if (psp.getColumnCount() != 3) {
        this.reportError(file.getPath() +":\n"
                        +"invalid number of columns (must be 3)");
        return false;           /* check the number of columns */
      }                         /* and abort if it is not 3 */
      psp.autoType();           /* determine the column types */
      Column c0 = psp.getColumn(0);
      Column c1 = psp.getColumn(1);
      Column c2 = psp.getColumn(2);
      if ((!(c0.getType() instanceof IntegerType))
      ||  (!(c1.getType() instanceof IntegerType))
      ||  (!(c2.getType() instanceof MetricType))) {
        this.reportError(file.getPath() +":\ninvalid column type"
                        +" (must be int, int, real)");
        return false;           /* check the types of the columns */
      }                         /* (int size, int supp, real count) */
      int min, max, m;          /* min./max. of size and support */
      IntegerType it;           /* to get the range of values */
      it  = (IntegerType)c0.getType();
      min = (int)it.getMin();   /* get range of values for */
      max = (int)it.getMax();   /* size and support column */
      it  = (IntegerType)c1.getType();
      m   = (int)it.getMin(); if (m < min) min = m;
      m   = (int)it.getMax(); if (m > max) max = m;
      NominalType nt = new NominalType();
      for (int i = min; i <= max; i++)  /* create nominal type with */
        nt.addValue(String.valueOf(i)); /* as many values as either */
      c0.setType(nt); c1.setType(nt);   /* attribute and set it */
      psp.renameColumn(0, "size");      /* give the three columns */
      psp.renameColumn(1, "support");   /* their proper names and */
      psp.renameColumn(2, "count");     /* create log(count) column */
      double[] d2 = (double[])c2.getData();
      double[] d3 = new double[d2.length];
      RealType rt = new RealType();
      for (int i = psp.getRowCount(); --i >= 0; ) {
        if (d2[i] > 0) rt.addValue(d3[i] = Math.log10(d2[i]));
        else           rt.setNull (d3,i);
      }                         /* scale counts logarithmically */
      psp.addColumn(new Column("log(count)", rt, d3));
      TableView tv = new TableView(this, TableView.ALL_ITEMS);
      tv.setTable(psp);         /* create a viewer for the table, */
      tv.setFormat(this.format);/* set the pattern spectrum table, */
      tv.setMode(0);            /* copy the data format and */
      tv.setVisible(true);      /* show the table viewer */
      t = System.currentTimeMillis() -t;
      System.err.print("[" +psp.getColumnCount() +" column(s), ");
      System.err.print(     psp.getRowCount()    +" row(s)] done");
      System.err.println(" [" +(t/(float)1000.0) +"s]."); }
     catch (FileNotFoundException e) {
      this.reportError(e.getMessage()); return false; }
    catch (IOException e) {     /* catch and report an i/o error */
      this.reportError(file.getPath() +":\n" +e.getMessage());
      return false;
    }
    return true;                /* return 'loading successful' */
  }  /* showSpectrum() */

  /*------------------------------------------------------------------*/
  /** Show found patterns/item sets.
   *  @param  txt the text field containing the file name
   *  @since  2013.11.28 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  protected void showPatterns (JTextField txt)
  { this.showPatterns(new File(txt.getText())); }

  /*------------------------------------------------------------------*/
  /** Show found patterns/item sets.
   *  @param  file the file to load the spike trains
   *  @since  2013.11.28 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  protected void showPatterns (File file)
  {                             /* --- show a pattern set */
    PatternView view = new PatternView(this, PatternView.FILE_ITEMS);
    if (!view.loadPatterns(file)) return;
    view.setVisible(true); view.toFront();
  }  /* showPatterns() */

  /*------------------------------------------------------------------*/
  /** Report an I/O error.
   *  @param  msg the error message to report
   *  @since  2013.11.27 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  private void reportError (String msg)
  {                             /* --- report an i/o error */
    this.status.setText(msg);   /* set message in status bar, */
    System.err.println();       /* print message to stderr, */
    System.err.println(msg);    /* and show alert dialog box */
    JOptionPane.showMessageDialog(this, msg,
      "Error", JOptionPane.ERROR_MESSAGE);
  }  /* reportError() */

  /*------------------------------------------------------------------*/
  /** Get an executor for the CoCoNAD function
   *  @return the executor for the CoCoNAD function
   *  @since  2013.11.25 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  protected Executor createExecutor (int i)
  {                             /* --- get the object to execute */
    PSFnPSR  exe;               /* PSF+PSR executable object */
    TrainSet trains;            /* trains to analyze */
    double   beg, end;          /* beginning and end of event range */
    int      cnt;               /* number of surrogate data sets */
    int      surr;              /* surrogate data generation method */
    int      rand;              /* random number density function */
    double   sigma;             /* parameter for random function */
    double   delta;             /* block size for blocked perms. */
    double   alpha;             /* item probability dispersion factor */
    int      smpls;             /* num. of samples per item set size */
    int      seed;              /* seed for random number generator */
    int      trg;               /* target pattern type */
    int      osub, ssub;        /* target pattern subtypes */
    int      supp;              /* minimum support of an item set */
    double   width;             /* width of the time window */
    int      min, max;          /* size range of the item sets */
    int      psr;               /* pattern set reduction method */
    String   s;                 /* for text field evaluation */

    exe = new PSFnPSR(new ActionListener () {
            public void actionPerformed (ActionEvent e) {
              CoCoGUI.this.status.setText(e.getActionCommand()); } });
    exe.setTrains  (this.fn_tns.getText(),
                    this.format.getRecSeps(),
                    this.format.getFldSeps(),
                    this.format.getBlanks(),
                    this.format.getComment(),
                    this.format.getTrainMode());
    exe.setSpectrum(this.fn_psp.getText(),
                    this.pspsep.getText());
    exe.setOutput  (this.fn_out.getText(),
                    this.header.getText(),
                    this.itemsep.getText(),
                    this.outfmt.getText());

    try { beg = Double.parseDouble(this.beg.getText()); }
    catch (NumberFormatException e) { beg = Double.NEGATIVE_INFINITY; }
    try { end = Double.parseDouble(this.end.getText()); }
    catch (NumberFormatException e) { end = Double.POSITIVE_INFINITY; }
    exe.setRange(beg, end);     /* set the allowed event range */

    cnt  = ((Integer)this.surcnt.getValue()).intValue();
    surr = surcodes[this.surr.getSelectedIndex()];
    rand = pdfcodes[this.pdf.getSelectedIndex()];
    s = this.sigma.getText().trim();
    try { sigma = Double.parseDouble(s); }
    catch (NumberFormatException e) { sigma = Double.NEGATIVE_INFINITY;}
    s = this.delta.getText().trim();
    try { delta = Double.parseDouble(s); }
    catch (NumberFormatException e) { delta = Double.NEGATIVE_INFINITY;}
    s = this.seed.getText().trim();
    try { seed  = Integer.parseInt(s); }
    catch (NumberFormatException e) { seed = 0; }
    exe.setSurrogate(cnt, surr, rand, sigma, delta, seed);

    s = this.alpha.getText().trim();
    try { alpha = Double.parseDouble(s); }
    catch (NumberFormatException e) { alpha = 0.5; }
    smpls = ((Integer)this.smpls.getValue()).intValue();
    exe.setEstimate(cnt, alpha, smpls, seed);

    trg  = trgcodes[this.target.getSelectedIndex()];
    osub = subcodes[this.suborig.getSelectedIndex()];
    ssub = subcodes[this.subsurr.getSelectedIndex()];
    min  = ((Integer)this.minsize.getValue()).intValue();
    max  = ((Integer)this.maxsize.getValue()).intValue();
    supp = ((Integer)this.minsupp.getValue()).intValue();
    try { width = Double.parseDouble(this.width.getText()); }
    catch (NumberFormatException e) { width = Double.NEGATIVE_INFINITY;}
    exe.setCoCoNAD(trg, osub, ssub, width, supp, min, max);

    psr = psrcodes[this.psr.getSelectedIndex()];
    exe.setReduction(psr);      /* set pattern set reduction method */

    return new ObjExecutor(exe, this);
  }  /* createExecutor() */

  /*------------------------------------------------------------------*/
  /** Get the result message on successful termination.
   *  @return the result message
   *  @since  2007.02.10 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  protected String getResultMsg ()
  {                             /* --- get result message */
    PSFnPSR exe = (PSFnPSR)((ObjExecutor)this.executor).getObject();
    String  msg = exe.getResultMsg();
    this.status.setText(msg);   /* get and set the result message */
    return msg;                 /* and then return it */
  }  /* getResultMsg() */

  /*------------------------------------------------------------------*/
  /** Load a configuration file and set the input fields.
   *  @param  file the file to load the configuration from
   *  @since  2013.11.25 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  protected void loadConfig (File file)
  {                             /* --- load configuration file */
    if (file == null) {         /* if no file name is given */
      JFileChooser c = this.getFileChooser();
      c.setDialogTitle("Load Configuration...");
      int r = c.showOpenDialog(this);
      if (r != JFileChooser.APPROVE_OPTION) return;
      file = c.getSelectedFile();
    }                           /* let the user choose a file */
    try {                       /* open the configuration file */
      FileReader reader = new FileReader(file);
      /* --- Format --- */
      this.format.loadConfig(reader);
      /* --- text parameters --- */
      /* --- Files --- */
      this.fn_tns.setText (this.readLine(reader));
      this.fn_psp.setText (this.readLine(reader));
      this.fn_out.setText (this.readLine(reader));
      this.beg.setText    (this.readLine(reader));
      this.end.setText    (this.readLine(reader));
      /* --- CoCoNAD --- */
      this.width.setText  (this.readLine(reader));
      /* --- PSF+PSR --- */
      this.sigma.setText  (this.readLine(reader));
      this.seed.setText   (this.readLine(reader));
      /* --- Output --- */
      this.pspsep.setText (this.readLine(reader));
      this.header.setText (this.readLine(reader));
      this.itemsep.setText(this.readLine(reader));
      this.outfmt.setText (this.readLine(reader));
      /* --- flags & indices --- */
      /* --- CoCoNAD --- */
      this.target.setSelectedIndex(this.readInt(reader));
      this.suborig.setSelectedIndex(this.readInt(reader));
      this.subsurr.setSelectedIndex(this.readInt(reader));
      this.minsize.setValue(Integer.valueOf(this.readInt(reader)));
      this.maxsize.setValue(Integer.valueOf(this.readInt(reader)));
      this.minsupp.setValue(Integer.valueOf(this.readInt(reader)));
      /* --- PSF+PSR --- */
      this.surr.setSelectedIndex(this.readInt(reader));
      this.pdf.setSelectedIndex (this.readInt(reader));
      this.psr.setSelectedIndex (this.readInt(reader));
      reader.close(); }         /* read the configuration values */
    catch (IOException e) {     /* and close the input file */
      JOptionPane.showMessageDialog(this,
        "Error reading configuration file:\n" +e.getMessage(),
        "Error", JOptionPane.ERROR_MESSAGE);
    }                           /* check for successful reading */
    this.status.setText("configuration loaded: " +file.getName());
  }  /* loadConfig() */

  /*------------------------------------------------------------------*/
  /** Save a configuration file
   *  @param  file the file to save the current configuration to
   *  @since  2013.11.25 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  protected void saveConfig (File file)
  {                             /* --- save configuration file */
    if (file == null) {         /* if no file name is given, */
      JFileChooser c = this.getFileChooser();
      c.setDialogTitle("Save Configuration...");
      int r = c.showOpenDialog(this);
      if (r != JFileChooser.APPROVE_OPTION) return;
      file = c.getSelectedFile();
    }                           /* let the user choose a file */
    try {                       /* open the configuration file */
      FileWriter writer = new FileWriter(file);
      /* --- Format --- */
      this.format.saveConfig(writer);
      /* --- text parameters --- */
      /* --- Files --- */
      writer.write(this.fn_tns.getText());  writer.write('\n');
      writer.write(this.fn_psp.getText());  writer.write('\n');
      writer.write(this.fn_out.getText());  writer.write('\n');
      writer.write(this.beg.getText());     writer.write('\n');
      writer.write(this.end.getText());     writer.write('\n');
      /* --- CoCoNAD --- */
      writer.write(this.width.getText());   writer.write('\n');
      /* --- PSF+PSR --- */
      writer.write(this.sigma.getText());   writer.write('\n');
      writer.write(this.seed.getText());    writer.write('\n');
      /* --- Output --- */
      writer.write(this.pspsep.getText());  writer.write('\n');
      writer.write(this.header.getText());  writer.write('\n');
      writer.write(this.itemsep.getText()); writer.write('\n');
      writer.write(this.outfmt.getText());  writer.write('\n');
      /* --- flags & indices --- */
      /* --- CoCoNAD --- */
      writer.write(this.target.getSelectedIndex()  +",");
      writer.write(this.suborig.getSelectedIndex() +",");
      writer.write(this.subsurr.getSelectedIndex() +",");
      writer.write(((Integer)this.minsize.getValue()).intValue() +",");
      writer.write(((Integer)this.maxsize.getValue()).intValue() +",");
      writer.write(((Integer)this.minsupp.getValue()).intValue() +",");
      /* --- PSF+PSR --- */
      writer.write(this.surr.getSelectedIndex() +",");
      writer.write(this.pdf.getSelectedIndex() +",");
      writer.write(this.psr.getSelectedIndex() +",");
      writer.write('\n');       /* write the configuration values */
      writer.close(); }         /* and close the output file */
    catch (IOException e) {     /* check for successful writing */
      JOptionPane.showMessageDialog(this,
        "Error writing configuration file:\n" +e.getMessage(),
        "Error", JOptionPane.ERROR_MESSAGE);
    }                           /* show a status message */
    this.status.setText("configuration saved: " +file.getName());
  }  /* saveConfig() */

  /*------------------------------------------------------------------*/
  /** Main function for command line invocation.
   *  @param  args the command line arguments
   *  @since  2013.11.25 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public static void main (String[] args)
  {                             /* --- main function */
    CoCoGUI gui = new CoCoGUI();
    if (args.length > 0)        /* load configuration if necessary */
      gui.loadConfig(new File(args[0]));
    gui.setVisible(true);       /* show the user interface */
  }  /* main() */

}  /* class CoCoGUI */
