/*----------------------------------------------------------------------
  File    : PSFnPSR.java
  Contents: Pattern Spectrum Filtering and Pattern Set Reduction
  Author  : Christian Borgelt
  History : 2013.12.10 file created
            2013.12.11 extraction of decision border added
            2013.12.12 pattern set reduction added
            2014.03.07 pattern spectrum estimation added
            2014.05.12 pattern spectrum border passed to coconad
            2014.06.10 use of decision border in reduce() corrected
            2014.06.17 check for no reduction added to exec()
            2014.08.11 item sequence mining added (respect item order)
            2014.10.03 adapted to new TrainSet function getAllTrains()
            2015.03.05 handling of user abort request improved
            2015.08.11 bug with empty pattern spectrum file fixed
            2015.08.12 adapted to modified pattern set reduction
            2015.08.13 function JNICoCo.patspec() renamed to genpsp()
            2016.11.04 adapted to fim modules (Pattern, PatternSet etc.)
----------------------------------------------------------------------*/
package coconad;

import java.util.Arrays;
import java.io.IOException;
import java.io.File;
import java.io.FileWriter;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import javax.swing.Timer;

import util.TableReader;
import util.Executable;
import util.IdMap;
import util.ASCIICoder;
import table.Table;
import table.Column;
import table.IntegerType;
import table.MetricType;
import fim.Pattern;
import fim.PatternSet;

/*--------------------------------------------------------------------*/
/** Class for Pattern Spectrum Filtering and Pattern Set Reduction.
 *  Provides an executable object that can be run as a thread.
 *  @author Christian Borgelt
 *  @since  2013.12.10 */
/*--------------------------------------------------------------------*/
public class PSFnPSR implements Executable {

  /* --- target pattern types --- */
  /** target pattern type: all frequent item sets */
  public static final int ITEMSET     = JNICoCo.ITEMSET;
  /** target pattern type: all frequent item sets */
  public static final int PERMUTATION = JNICoCo.PERMUTATION;
  /** target pattern type: all closed (frequent) item sets */
  public static final int SEQUENCE    = JNICoCo.SEQUENCE;

  /* --- target pattern subtypes --- */
  /** target pattern subtype: all frequent item sets */
  public static final int FREQUENT    = JNICoCo.FREQUENT;
  /** target pattern subtype: all frequent item sets */
  public static final int ALL         = JNICoCo.ALL;
  /** target pattern subtype: all closed (frequent) item sets */
  public static final int CLOSED      = JNICoCo.CLOSED;
  /** target pattern subtype: all maximal (frequent) item sets */
  public static final int MAXIMAL     = JNICoCo.MAXIMAL;

  /* --- CoCoNAD algorithm variants --- */
  /** algorithm variant: basic version (low memory consumption, slow) */
  public static final int BASIC       = JNICoCo.BASIC;
  /** algorithm variant: refined version (spike and train filtering) */
  public static final int FILTER      = JNICoCo.FILTER;
  /** algorithm variant: refined version (closed/maximal repository) */
  public static final int CMREPO      = JNICoCo.CMREPO;

  /* --- surrogate data generation methods --- */
  /** surrogate method: none (no pattern spectrum filtering) */
  public static final int NONE        = -3;
  /** surrogate method: read pattern spectrum from file */
  public static final int READ        = -2;
  /** surrogate method: estimate pattern spectrum (no surrogates) */
  public static final int ESTIMATE    = -1;
  /** surrogate method: identity (keep original data) */
  public static final int IDENTITY    = JNICoCo.IDENTITY;
  /** surrogate method: event/spike time rendomization */
  public static final int RANDOM      = JNICoCo.RANDOM;
  /** surrogate method: event/spike dithering/displacement */
  public static final int DITHER      = JNICoCo.DITHER;
  /** surrogate method: train shifting/dithering */
  public static final int SHIFT       = JNICoCo.SHIFT;
  /** surrogate method: sampling from kernel estimate */
  public static final int KERNEST     = JNICoCo.KERNEST;
  /** surrogate method: dithered event/spike permutation */
  public static final int PERMUTE     = JNICoCo.PERMUTE;
  /** surrogate method: dithered blocked permutations */
  public static final int BLOCK       = JNICoCo.BLOCK;

  /* --- random number generation methods --- */
  /** random number method: uniform density
   *  (identical to <code>RECT</code> */
  public static final int UNIFORM  = JNICoCo.UNIFORM;
  /** random number method: rectangular density
   *  (identical to <code>UNIFORM</code> */
  public static final int RECT     = JNICoCo.RECT;
  /** random number method: triangular density */
  public static final int TRIANG   = JNICoCo.TRIANG;
  /** random number method: Gaussian density
   *  (identical to <code>NORMAL</code> */
  public static final int GAUSS    = JNICoCo.GAUSS;
  /** random number method: normal density
   *  (identical to <code>GAUSS</code> */
  public static final int NORMAL   = JNICoCo.NORMAL;

  /*------------------------------------------------------------------*/
  /* Instance Variables                                               */
  /*------------------------------------------------------------------*/
  /** the progress reporter */
  private ActionListener progrep;
  /** the name of the trains file */
  private String         fn_tns;
  /** the name of the pattern spectrum file */
  private String         fn_psp;
  /** the name of the output/pattern file */
  private String         fn_out;
  /** the record  separators */
  private String         recseps;
  /** the field   separators */
  private String         fldseps;
  /** the blank   characters */
  private String         blanks;
  /** the comment characters */
  private String         comment;
  /** the input record mode (e.g. TrainSet.ITEM_TIME) */
  private int            mode;
  /** the pattern spectrum separator (characters) */
  private String         pspsep;
  /** the header string for pattern/item set output */
  private String         header;
  /** the item separator for pattern/item set output */
  private String         itemsep;
  /** the support output format */
  private String         outfmt;
  /** the set of (spike) trains to analyze */
  private TrainSet       trains;
  /** the beginning of allowed span of events */
  private double         beg;
  /** the end       of allowed span of events */
  private double         end;
  /** the number of surrogate data sets to generate */
  private int            surcnt;
  /** the surrogate data generation method (<code>NONE</code>
   *  <code>READ</code>, <code>ESTIMATE</code>, <code>IDENTITY</code>,
   *  <code>RANDOM</code>, <code>DITHER</code>, <code>SHIFT</code>,
   *  <code>KERNEST</code> or <code>PERMUTE</code>) */
  private int            surr;
  /** the random number density function (for displacements)
   *  (<code>UNIFORM</code>, <code>RECT</code>, <code>TRIANG</code>,
   *   <code>GAUSS</code> or <code>NORMAL</code>) */
  private int            rand;
  /** the standard deviation or half the base width
   *  of the rectangular/triangular density function */
  private double         sigma;
  /** the block size (time interval) for blocked permutations */
  private double         delta;
  /** the equivalent number of surrogate data sets */
  private int            equiv;
  /** the item probability dispersion factor */
  private double         alpha;
  /** the number of samples per item set size */
  private int            smpls;
  /** the seed value for random number generator */
  private int            seed;
  /** the target pattern type (<code>ITEMSET</code>,
   *  <code>PERMUTATION</code> or <code>SEQUENCE</code>) */
  private int            target;
  /** the target pattern subtype for the original data,
   *  (<code>FREQUENT</code>, <code>ALL</code>, <code>CLOSED</code>
   *  or <code>MAXIMAL</code>) */
  private int            suborig;
  /** the target pattern subtype for the surrogates,
   *  (<code>FREQUENT</code>, <code>ALL</code>, <code>CLOSED</code>
   *  or <code>MAXIMAL</code>) */
  private int            subsurr;
  /** the minimum support of an item set */
  private int            supp;
  /** the width of the time window/maximum time span */
  private double         width;
  /** the minimum number of items per item set */
  private int            zmin;
  /** the maximum number of items per item set */
  private int            zmax;
  /** the algorithm variant to use (<code>BASIC</code>,
   *  <code>FILTER</code> or <code>CMREPO</code>) */
  private int            algo;
  /** the processing mode to use (<code>JNICoCo.NONE</code> or flags
   *  <code>JNICoCo.NOPERFECT</code> or <code>JNICoCo.NOSORT</code>) */
  private int            pmod;
  /** the number of cpus/threads to use */
  private int            cpus;
  /** the control array (currently only progress indicator) */
  private int[]          ctrl;
  /** the pattern set reduction method */
  private int            psr;
  /** the result of pattern spectrum generation */
  private Object[]       spectrum;
  /** the decision border derived from the pattern spectrum */
  private int[]          border;
  /** the result of the pattern mining with CoCoNAD */
  private PatternSet     patterns;
  /** the number of found patterns */
  private int            patcnt;
  /** a buffer for pattern intersections */
  private int[]          buf;
  /** the execution time of the computations */
  private double         exetime;
  /** the stop flag for the execution */
  private boolean        stop;
  /** timer for progress messages */
  private Timer          timer;

  /*------------------------------------------------------------------*/
  /** Create an object for pattern spectrum filtering
   *  and pattern set reduction.
   *  @since  2013.12.10 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public PSFnPSR ()
  { this(null); }

  /*------------------------------------------------------------------*/
  /** Create an object for pattern spectrum filtering
   *  and pattern set reduction.
   *  @param  progrep  the progress reporter
   *  @since  2013.12.10 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public PSFnPSR (ActionListener progrep)
  {                             /* --- create PSF+PSR object */
    this.progrep  = progrep;
    this.fn_tns   = null;
    this.fn_psp   = null;
    this.fn_out   = null;
    this.recseps  = "\n";
    this.fldseps  = " ,\t";
    this.blanks   = " \t\r";
    this.comment  = "#";
    this.mode     = TrainSet.ITEM_TIME;
    this.pspsep   = " ";
    this.header   = "";
    this.itemsep  = " ";
    this.outfmt   = " (%d)";
    this.trains   = null;       /* clear the train set */
    this.beg      = Double.NEGATIVE_INFINITY;
    this.end      = Double.POSITIVE_INFINITY;
    this.surcnt   = 1000;
    this.surr     = PERMUTE;    /* set reasonable default values */
    this.rand     = UNIFORM;
    this.sigma    = 0.005;
    this.delta    = 0.03;
    this.equiv    = 1000;
    this.alpha    = 0.5;
    this.smpls    = 1000;
    this.seed     = 0;
    this.target   = ITEMSET;
    this.suborig  = CLOSED;
    this.subsurr  = FREQUENT;
    this.supp     = 2;
    this.width    = 0.003;
    this.zmin     = 2;
    this.zmax     = -1;
    this.pmod     = JNICoCo.NONE;
    this.algo     = CMREPO;
    this.cpus     = 0;
    this.ctrl     = new int[1];
    this.psr      = PatternSet.COVER1;
    this.spectrum = null;
    this.border   = null;
    this.patterns = null;
    this.patcnt   = 0;
    this.stop     = false;
  }  /* PSFnPSR() */

  /*------------------------------------------------------------------*/
  /** Set the name of the train set file.
   *  @param  fn_tns the name of the train set file
   *  @since  2013.12.10 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public void setTrains (String fn_tns)
  { this.setTrains(fn_tns, "\\n", " ,\\t", " \\t\\r", "#", 0); }

  /*------------------------------------------------------------------*/
  /** Set the name of the train set file and the record contents mode.
   *  @param  fn_tns the name of the train set file
   *  @param  mode   the record contents mode
   *                 (<code>TrainSet.ITEM_POINT</code>,
   *                  <code>TrainSet.POINT_ITEM</code>,
   *                  <code>TrainSet.ITEM_TRAIN</code> or
   *                  <code>TrainSet.PURE_TRAIN</code>)
   *  @since  2013.12.10 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public void setTrains (String fn_tns, int mode)
  { this.setTrains(fn_tns, "\\n", " ,\\t", " \\t\\r", "#", mode); }

  /*------------------------------------------------------------------*/
  /** Set the name of the train set file, the structure characters,
   *  and the record contents mode.
   *  @param  fn_tns  the name of the train set file
   *  @param  recseps the record  separators
   *  @param  fldseps the field   separators
   *  @param  blanks  the blank   characters
   *  @param  comment the comment characters
   *  @param  mode    the record contents mode
   *                  (<code>TrainSet.ITEM_POINT</code>,
   *                   <code>TrainSet.POINT_ITEM</code>,
   *                   <code>TrainSet.ITEM_TRAIN</code> or
   *                   <code>TrainSet.PURE_TRAIN</code>)
   *  @since  2013.12.10 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public void setTrains (String fn_tns, String recseps, String fldseps,
                         String blanks, String comment, int mode)
  {                             /* --- set the trains file name */
    this.fn_tns  = fn_tns;  this.mode    = mode;
    this.recseps = recseps; this.fldseps = fldseps;
    this.blanks  = blanks;  this.comment = comment;
  }  /* setTrains() */

  /*------------------------------------------------------------------*/
  /** Set the train set to analyze.
   *  @param  trains  the train set to analyze
   *  @since  2013.12.10 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public void setTrains (TrainSet trains)
  { this.trains = trains; }

  /*------------------------------------------------------------------*/
  /** Set the name of the pattern spectrum file.
   *  @param  fn_psp  the name of the pattern spectrum file
   *  @since  2013.12.10 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public void setSpectrum (String fn_psp)
  { this.setSpectrum(fn_psp, " "); }

  /*------------------------------------------------------------------*/
  /** Set the name of the pattern spectrum file.
   *  @param  fn_psp the name of the pattern spectrum file
   *  @param  pspsep the field separators for the pattern spectrum
   *  @since  2013.12.10 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public void setSpectrum (String fn_psp, String pspsep)
  {                             /* --- set the spectrum file name */
    this.fn_psp = fn_psp; this.pspsep = pspsep;
  }  /* setSpectrum() */

  /*------------------------------------------------------------------*/
  /** Set the pattern spectrum.
   *  @param  spectrum  the pattern spectrum as a 3 column table
   *  @since  2013.12.10 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public void setSpectrum (Table spectrum)
  {                             /* --- set the pattern spectrum */
    if (spectrum == null) { this.spectrum = null; return; }
    this.spectrum    = new Object[3];
    this.spectrum[0] = spectrum.getColumn(0).getData();
    this.spectrum[1] = spectrum.getColumn(1).getData();
    this.spectrum[2] = spectrum.getColumn(2).getData();
  }  /* setSpectrum() */

  /*------------------------------------------------------------------*/
  /** Set the pattern output file name.
   *  @param  fn_out  the pattern output file name
   *  @since  2013.12.10 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public void setOutput (String fn_out)
  { this.setOutput(fn_out, "", " ", " (%d)"); }

  /*------------------------------------------------------------------*/
  /** Set the pattern output file name and the format strings.
   *  @param  fn_out  the pattern output file name
   *  @param  header  the record header (output before each pattern)
   *  @param  itemsep the item separator (output between patterns)
   *  @param  outfmt  the support output format
   *  @since  2013.12.10 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public void setOutput (String fn_out,  String header,
                         String itemsep, String outfmt)
  {                             /* --- set the output file name */
    this.fn_out  = fn_out;  this.header = header;
    this.itemsep = itemsep; this.outfmt = outfmt;
  }  /* setOutput() */

  /*------------------------------------------------------------------*/
  /** Set the (allowed) range of points/spike times.
   *  @param  beg  the beginning of the point range
   *  @param  end  the end       of the point range
   *  @since  2013.12.10 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public void setRange (double beg, double end)
  { this.beg = beg; this.end = end; }

  /*------------------------------------------------------------------*/
  /** Set the surrogate data parameters.
   *  @param  cnt   the number of surrogate data sets to generate
   *  @param  surr  the surrgate data generation method
   *                (e.g. <code>DITHER</code> or <code>PERMUTE</code>)
   *  @param  rand  the random number density function type
   *                (e.g. <code>UNIFORM</code> or <code>NORMAL</code>)
   *  @param  sigma the random number density function parameter
   *  @param  delta the block size for blocked permutations
   *  @since  2013.12.10 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public void setSurrogate (int cnt, int surr, int rand,
                            double sigma, double delta)
  {                             /* --- set surrogate parameters */
    this.surcnt = cnt;   this.surr  = surr;
    this.rand   = rand;  this.sigma = sigma;
    this.delta  = delta; this.seed  = 0;
  }  /* setSurrogate() */

  /*------------------------------------------------------------------*/
  /** Set the surrogate data parameters.
   *  @param  cnt   the number of surrogate data sets to generate
   *  @param  surr  the surrgate data generation method
   *  @param  rand  the random number density function type
   *  @param  sigma the random number density function parameter
   *  @param  delta the block size for blocked permutations
   *  @param  seed  the seed for the random number generator
   *  @since  2013.12.10 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public void setSurrogate (int cnt, int surr, int rand,
                            double sigma, double delta, int seed)
  {                             /* --- set surrogate parameters */
    this.surcnt = cnt;   this.surr  = surr;
    this.rand   = rand;  this.sigma = sigma;
    this.delta  = delta; this.seed  = seed;
  }  /* setSurrogate() */

  /*------------------------------------------------------------------*/
  /** Set the pattern spectrum estimation parameters.
   *  @param  equiv the equivalent number of surrogate data sets
   *  @param  alpha the item probability dispersion factor
   *  @param  smpls the number of samples per item set size
   *  @since  2014.03.07 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public void setEstimate (int equiv, double alpha, int smpls)
  {                             /* --- set estimation parameters */
    this.equiv = equiv; this.alpha = alpha;
    this.smpls = smpls; this.seed  = 0;
  }  /* setEstimate() */

  /*------------------------------------------------------------------*/
  /** Set the pattern spectrum estimation parameters.
   *  @param  equiv the equivalent number of surrogate data sets
   *  @param  alpha the item probability dispersion factor
   *  @param  smpls the number of samples per item set size
   *  @param  seed  the seed for the random number generator
   *  @since  2014.03.07 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public void setEstimate (int equiv, double alpha, int smpls, int seed)
  {                             /* --- set estimation parameters */
    this.equiv = equiv; this.alpha = alpha;
    this.smpls = smpls; this.seed  = seed;
  }  /* setEstimate() */

  /*------------------------------------------------------------------*/
  /** Set the CoCoNAD parameters.
   *  @param  target  the target pattern type
   *                  (<code>ITEMSET</code>, <code>PERMUTATION</code>,
   *                   or <code>SEQUENCE</code>)
   *  @param  suborig the target pattern subtype for the original data
   *                  (<code>FREQUENT</code>, <code>SETS</code>,
   *                   <code>CLOSED</code> or <code>MAXIMAL</code>)
   *  @param  subsurr the target pattern subtype for the surrogate data
   *                  (<code>FREQUENT</code>, <code>SETS</code>,
   *                   <code>CLOSED</code> or <code>MAXIMAL</code>)
   *  @param  supp    the minimum support of an item set
   *  @param  width   the window width/maximum distance
   *  @param  zmin    the minimal size of an item set
   *  @param  zmax    the maximal size of an item set
   *  @since  2013.12.10 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public void setCoCoNAD (int target, int suborig, int subsurr,
                          double width, int supp, int zmin, int zmax)
  {                             /* --- set CoCoNAD parameters */
    this.target  = target;
    this.suborig = suborig;
    this.subsurr = subsurr;
    this.width   = width;  this.supp = supp;
    this.zmin    = zmin;   this.zmax = zmax;
    this.algo    = CMREPO; this.pmod = JNICoCo.NONE;
  }  /* setCoCoNAD() */

  /*------------------------------------------------------------------*/
  /** Set the CoCoNAD parameters.
   *  @param  target  the target pattern type
   *                  (<code>ITEMSET</code>, <code>PERMUTATION</code>,
   *                   or <code>SEQUENCE</code>)
   *  @param  suborig the target pattern subtype for the original data
   *                  (<code>FREQUENT</code>, <code>SETS</code>,
   *                   <code>CLOSED</code> or <code>MAXIMAL</code>)
   *  @param  subsurr the target pattern subtype for the surrogate data
   *                  (<code>FREQUENT</code>, <code>SETS</code>,
   *                   <code>CLOSED</code> or <code>MAXIMAL</code>)
   *  @param  supp    the minimum support of an item set
   *  @param  width   the window width/maximum distance
   *  @param  zmin    the minimal size of an item set
   *  @param  zmax    the maximal size of an item set
   *  @param  algo    the CoCoNAD algorithm variant to use
   *                  (<code>BASIC</code>, <code>FILTER</code> or
   *                   <code>CMREPO</code>)
   *  @param  mode    the processign mode to use (<code>NONE</code>
   *                  or flags <code>NOPERFECT</code>
   *                  or <code>NOSORT</code>)
   *  @since  2013.12.10 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public void setCoCoNAD (int target, int suborig, int subsurr,
                          double width, int supp, int zmin, int zmax,
                          int algo, int mode)
  {                             /* --- set CoCoNAD parameters */
    this.target  = target;
    this.suborig = suborig;
    this.subsurr = subsurr;
    this.width   = width; this.supp = supp;
    this.zmin    = zmin;  this.zmax = zmax;
    this.algo    = algo;  this.pmod = mode;
  }  /* setCoCoNAD() */

  /*------------------------------------------------------------------*/
  /** Set the pattern set reduction method.
   *  @param  psr  the pattern set reduction method
   *  @since  2013.12.10 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public void setReduction (int psr)
  { this.psr = psr; }

  /*------------------------------------------------------------------*/
  /** Set the number of CPUs/threads to use.
   *  @param  cpus  the number of CPUs/threads to use
   *                (0: determine a suitable number automatically)
   *  @since  2013.12.10 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public void setCPUs (int cpus)
  { this.cpus = cpus; }

  /*------------------------------------------------------------------*/
  /** Report progress of the execution.
   *  @param  msg  the message to report
   *  @since  2013.12.12 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  private void report (String msg)
  {                             /* report execution progress */
    if (this.progrep != null)   /* check for a progress reporter */
      this.progrep.actionPerformed(new ActionEvent(this, 0, msg));
  }  /* report() */

  /*------------------------------------------------------------------*/
  /** Execute pattern spectrum filtering and pattern set reduction.
   *  @since  2013.12.10 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public void exec () throws IOException
  {                             /* --- execute PSF+PSR */
    int         i, k;           /* loop variables */
    long        t, tall;        /* for time measurements */
    TableReader reader;         /* reader for input files */
    Table       psp;            /* pattern spectrum as a table */
    Column      c0, c1, c2;     /* columns of the pattern spectrum */
    FileWriter  writer;         /* writer for patterns/spectrum */
    char        sep;            /* field separator for writing */
    int[]       sizes;          /* sizes          of pattern spectrum */
    int[]       supps;          /* support values of pattern spectrum */
    double[]    frqs;           /* frequencies    of pattern spectrum */
    String      isep, fmt;      /* item separator and support format */
    Pattern     p;              /* to traverse the patterns */

    JNICoCo.abort(0);           /* clear the abort flag */
    tall = System.currentTimeMillis();
    this.exetime = 0;           /* start the overall timer */

    /* --- read trains --- */
    if (this.trains == null) {  /* if there is no train set */
      if (this.fn_tns.length() <= 0)  /* check for a file name */
        throw new IOException("Need a (spike) trains file name.");
      this.report("Reading " +this.fn_tns +" ...");
      System.err.print("reading " +this.fn_tns +" ... ");
      t = System.currentTimeMillis();
      reader = new TableReader(new File(this.fn_tns));
      reader.setCharsCoded(this.recseps, this.fldseps,
                           this.blanks,  "", this.comment);
      this.trains = TrainSet.parse(null, reader, this.mode);
      reader.close();           /* read trains from the given file */
      this.trains.filter(this.beg, this.end);
      this.trains.pack();       /* filter and pack the trains */
      t = System.currentTimeMillis() -t;
      System.err.print("[" +this.trains.getSize() +" trains(s)] done");
      System.err.println(" [" +(t/(float)1000.0) +"s].");
    }                           /* print a log message */
    if (this.stop) return;      /* check for abort */

    /* --- read pattern spectrum --- */
    if ((this.spectrum == null) /* if there is no pattern spectrum */
    &&  (this.surr == READ)) {  /* and to read one from a file */
      if ((this.fn_psp == null) /* check for a file name */
      ||  (this.fn_psp.length() <= 0))
        throw new IOException("Need a pattern spectrum file name.");
      this.report("Reading " +this.fn_psp +" ...");
      System.err.print("reading " +this.fn_psp +" ... ");
      t = System.currentTimeMillis();
      reader = new TableReader(new File(this.fn_psp));
      reader.setCharsCoded("\\n", this.pspsep, " ", "", this.comment);
      psp    = new Table(this.fn_tns);
      psp.read(reader, Table.NOHEADER);
      reader.close();           /* read the pattern spectrum */
      System.err.print("rows: "+String.valueOf(psp.getRowCount()));
      if (psp.getColumnCount() != 3)
        throw new IOException(this.fn_psp +":\n"
                  +"invalid number of columns (must be 3)");
      psp.autoType();         /* determine the column types */
      c0 = psp.getColumn(0);  /* get the columns and */
      c1 = psp.getColumn(1);  /* check their data types */
      c2 = psp.getColumn(2);
      if ((!(c0.getType() instanceof IntegerType))
      ||  (!(c1.getType() instanceof IntegerType))
      ||  (!(c2.getType() instanceof MetricType)))
        throw new IOException(this.fn_psp +":\ninvalid column type"
                  +" (must be int, int, real)");
      this.setSpectrum(psp);    /* set the read pattern spectrum */
      t = System.currentTimeMillis() -t;
      System.err.print("[" +psp.getRowCount() +" signature(s)] done");
      System.err.println(" [" +(t/(float)1000.0) +"s].");
    }                           /* print a log message */
    if (this.stop) return;      /* check for abort */

    /* --- estimate pattern spectrum --- */
    if ((this.spectrum == null) /* if there is no pattern spectrum */
    &&  (this.surr == ESTIMATE)) {          /* and to estimate one */
      this.report("Estimating pattern spectrum ... ");
      System.err.print("estimating pattern spectrum ... ");
      t = System.currentTimeMillis();
      this.spectrum = JNICoCo.estpsp(this.trains.getAllTrains(),
                        this.target|this.subsurr, this.width, this.supp,
                        this.zmin,  this.zmax, "#", this.equiv,
                        this.alpha, this.smpls, this.seed);
      k = ((int[])this.spectrum[0]).length;
      t = System.currentTimeMillis() -t;
      System.err.print("[" +k +" signature(s)] done");
      System.err.println(" [" +(t/(float)1000.0) +"s].");
    }                           /* print a log message */
    if (this.stop) return;      /* check for abort */

    /* --- analyze surrogate data --- */
    if ((this.spectrum == null) /* if there is no pattern spectrum */
    &&  (this.surr >= IDENTITY)) {          /* and to generate one */
      this.report("Analyzing surrogate data ... 0");
      System.err.print("analyzing surrogate data ... ");
      if (this.progrep != null){/* if to report the progress */
        timer = new Timer(500, new ActionListener () {
          public void actionPerformed (ActionEvent e) {
            PSFnPSR.this.report("Analyzing surrogate data ... "
                                +PSFnPSR.this.ctrl[0]);
          } } );                /* update the progress reporter */
        timer.start();          /* start a timer that triggers */
      }                         /* progress report updates */
      t = System.currentTimeMillis();
      this.spectrum = JNICoCo.genpsp(this.trains.getAllTrains(),
                        this.target|this.subsurr, this.width, this.supp,
                        this.zmin,  this.zmax, "#", this.algo,this.pmod,
                        this.surcnt,this.beg,   this.end,   this.surr,
                        this.rand,  this.sigma, this.delta, this.seed,
                        this.cpus,  this.ctrl);
      k = ((int[])this.spectrum[0]).length;
      t = System.currentTimeMillis() -t;
      System.err.print("[" +k +" signature(s)] done");
      System.err.println(" [" +(t/(float)1000.0) +"s].");
      if (this.progrep != null) /* if to report the progress, */
        timer.stop();           /* stop the timer */
    }
    if (this.stop) return;      /* check for abort */

    /* --- write pattern spectrum --- */
    if ((this.fn_psp != null)   /* if to write the pattern spectrum */
    &&  (this.surr >= ESTIMATE)) {         /* and a one was created */
      if (this.fn_psp.length() <= 0)
        throw new IOException("Need a pattern spectrum file name.");
      System.err.print("writing " +this.fn_psp +" ... ");
      t = System.currentTimeMillis();
      writer = new FileWriter(new File(this.fn_psp));
      sep    = (this.pspsep.length() <= 0) ? ' '
             : ASCIICoder.decode(this.pspsep).charAt(0);
      sizes  = (int[])   this.spectrum[0];
      supps  = (int[])   this.spectrum[1];
      frqs   = (double[])this.spectrum[2];
      for (i = 0; i < sizes.length; i++) {
        writer.write(String.valueOf(sizes[i])); writer.write(sep);
        writer.write(String.valueOf(supps[i])); writer.write(sep);
        writer.write(String.valueOf(frqs [i])); writer.write('\n');
      }                       /* write the triplets and */
      writer.close();         /* finally close the writer */
      t = System.currentTimeMillis() -t;
      System.err.print("[" +sizes.length +" signature(s)] done");
      System.err.println(" [" +(t/(float)1000.0) +"s].");
    }                           /* print a log message */
    if (this.stop) return;      /* check for abort */

    /* --- analyzing original trains --- */
    this.report("Analyzing original data ...");
    System.err.print("analyzing original data ... ");
    t = System.currentTimeMillis();
    if (this.spectrum != null)  /* compute the decision border */
      this.border = JNICoCo.psp2bdr(this.spectrum);
    this.patterns = JNICoCo.coconad(this.trains,
                      this.target|this.suborig, this.width,
                      this.supp,  this.zmin,    this.zmax,
                      this.algo,  this.pmod,    this.border);
    this.patterns.sort();       /* mine patterns and sort them */
    this.patcnt = this.patterns.getCount();
    t = System.currentTimeMillis() -t;
    System.err.print("[" +this.patcnt +" pattern(s)] done");
    System.err.println(" [" +(t/(float)1000.0) +"s].");
    if (this.stop) return;      /* check for abort */

    /* --- pattern set reduction --- */
    if ((this.target == ITEMSET)  /* if not to mine sequences */
    &&  (this.surr   != IDENTITY) /* and to reduce pattern set */
    &&  (this.psr    != PatternSet.KEEP)) {
      this.report("Pattern set reduction ...");
      System.err.print("pattern set reduction ... ");
      t = System.currentTimeMillis();
      this.patcnt = this.patterns.reduce(this.psr, this.border, true);
      t = System.currentTimeMillis() -t;
      System.err.print("[" +this.patcnt +" pattern(s)] done");
      System.err.println(" [" +(t/(float)1000.0) +"s].");
    }                           /* print a log message */
    if (this.stop) return;      /* check for abort */

    /* --- report final result --- */
    if (this.fn_out != null) {  /* if to write an output file */
      if (this.fn_out.length() <= 0)
        throw new IOException("Need a pattern set file name.");
      this.report("Writing " +this.fn_out +" ...");
      System.err.print("writing " +this.fn_out +" ... ");
      t = System.currentTimeMillis();
      isep   = (this.itemsep.length() <= 0) ? " "
             : ASCIICoder.decode(this.itemsep);
      fmt    = (this.outfmt.length() <= 0) ? ""
             : ASCIICoder.decode(this.outfmt);
      writer = new FileWriter(new File(this.fn_out));
      this.patterns.write(writer, isep, fmt);
      writer.close();           /* write the pattern set */
      t = System.currentTimeMillis() -t;
      System.err.print("[" +this.patcnt +" pattern(s)] done");
      System.err.println(" [" +(t/(float)1000.0) +"s].");
    }                           /* print a log message */

    tall = System.currentTimeMillis() -tall;
    this.exetime = tall/(float)1000.0;
    this.report("Found "+this.patcnt+" pattern(s) in "
               +String.format("%.3f", this.exetime) +"s.");
  }  /* exec() */

  /*------------------------------------------------------------------*/
  /** Abort the computations (set an abort flag).
   *  @since  2013.12.10 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public void abort ()
  {                             /* --- abort computations */
    if (timer != null) timer.stop();
    JNICoCo.abort(-1);          /* stop timer, abort JNI functions */
    this.stop = true;           /* and set the stop flag */
  }  /* abort() */

  /*------------------------------------------------------------------*/
  /** Get the result message of the computations.
   *  @return the result message of the computations
   *  @since  2013.12.10 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public String getResultMsg ()
  { return "Found " +this.patcnt +" pattern(s) in "
           +String.format("%.3f", this.exetime) +"s."; }

}  /* class PatSpec */
